/*
 * $Id: qm_mopac.c,v 1.1 2005/09/02 00:30:56 lindahl Exp $
 * 
 *                This source code is part of
 * 
 *                 G   R   O   M   A   C   S
 * 
 *          GROningen MAchine for Chemical Simulations
 * 
 *                        VERSION 3.2.0
 * Written by David van der Spoel, Erik Lindahl, Berk Hess, and others.
 * Copyright (c) 1991-2000, University of Groningen, The Netherlands.
 * Copyright (c) 2001-2004, The GROMACS development team,
 * check out http://www.gromacs.org for more information.
 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * If you want to redistribute modifications, please consider that
 * scientific software is very special. Version control is crucial -
 * bugs must be traceable. We will be happy to consider code for
 * inclusion in the official distribution, but derived work must not
 * be called official GROMACS. Details are found in the README & COPYING
 * files - if they are missing, get the official version at www.gromacs.org.
 * 
 * To help us fund GROMACS development, we humbly ask that you cite
 * the papers on the package - you can find them in the top README file.
 * 
 * For more info, check our website at http://www.gromacs.org
 * 
 * And Hey:
 * GROwing Monsters And Cloning Shrimps
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <math.h>
#include "sysstuff.h"
#include "typedefs.h"
#include "macros.h"
#include "smalloc.h"
#include "assert.h"
#include "physics.h"
#include "macros.h"
#include "vec.h"
#include "force.h"
#include "invblock.h"
#include "confio.h"
#include "nsb.h"
#include "names.h"
#include "network.h"
#include "wnblist.h"
#include "pbc.h"
#include "ns.h"
#include "nrnb.h"
#include "bondf.h"
#include "mshift.h"
#include "txtdump.h"
#include "copyrite.h"
#include "qmmm.h"
#include <stdio.h>
#include <string.h>
#include "fatal.h"
#include "typedefs.h"
#include <stdlib.h>

/* DFTB interface routines */

#ifndef F77_FUNC
#define F77_FUNC(name,NAME) name ## _
#endif


void 
F77_FUNC(gmxcao,GMXCAO)(double *scftol,int *charge,int *nrqmat, int izp[], 
			double qmcrd[],int lmax[],char skf[], double *telec,
			int number[]);

void 
F77_FUNC(dodftb,DODFTB)(double *scftol,int *nrqmat,double qmcrd[],int *nrmmat,
			double mmcrd[],double mmchrg[],double qmgrad[],
			double mmgrad[], double *telec, double *energy);



void init_dftb(t_commrec *cr, t_QMrec *qm, t_MMrec *mm)
{
  int
    i,j,k=0,*izp=NULL,*izp_atnum=NULL,izp_nr=0,
    *lmax=NULL;
  double
    telec,
    *qmcrd =NULL;  
  bool
    unique;
  char
    *buf=NULL,*at1=NULL,*at2=NULL,*skf=NULL;
  FILE
    *tmpfile;
  int aantal;
  double
    scftol;

  telec=(double)qm->telec;
  snew(izp,qm->nrQMatoms);
  snew(qmcrd, 3*qm->nrQMatoms);

  //  aantal[0]=5;

  for(i=0;i<qm->nrQMatoms;i++){
    for (j=0;j<DIM;j++){
      qmcrd[3*i+j] = (double)qm->xQM[i][j]*10;
    }
  }
  izp_nr=1;
  snew(izp_atnum,izp_nr);

  /* check for unique atom types */
  izp_atnum[0]=qm->atomicnumberQM[0];
  for(i=1;i<qm->nrQMatoms;i++){
    unique = TRUE;
    for(j=0;j<izp_nr;j++){
      if(izp_atnum[j]==qm->atomicnumberQM[i]){
	unique = FALSE;
      }
    }
    if(unique){
      srenew(izp_atnum,izp_nr+1);
      izp_atnum[izp_nr++]=qm->atomicnumberQM[i];
      
    }
  }
  snew(izp,qm->nrQMatoms);
  for(i=0;i<qm->nrQMatoms;i++){
    for(j=0;j<izp_nr;j++){
      if(qm->atomicnumberQM[i]==izp_atnum[j]){
	izp[i] = j+1;
      }
    }
  }
  /* we now have izp, next is to create the lmax array */
  snew(lmax,izp_nr);
  for(i=0;i<izp_nr;i++){
    if(izp_atnum[i]<=2)
      lmax[i]=1;/*  first row elements */
    else if(izp_atnum[i]<=10)
      lmax[i]=2;/* second row elements */
    else if(izp_atnum[i]<=18)
      lmax[i]=3;/*  third row elements */
    else if(izp_atnum[i]<=36)
      lmax[i]=4;
    else if(izp_atnum[i]<=54)
      lmax[i]=5;
    else if(izp_atnum[i]<=86)
      lmax[i]=6;
    else if(izp_atnum[i]<=118)
      lmax[i]=7;
  }

  /* now deal with the slko files for the atoms in the QM subsystem
   */

  buf = getenv("SLKO_PATH");
  if(!buf)
    gmx_fatal(FARGS,"no $SLKO_PATH, where the slko files are located\n, you can obtain these files from www.dftb.org\n");
  
  /* now set the correct order of the slko files for the system at hand 
   */

  snew(skf,2048);
  for(i=0;i<izp_nr;i++){

    snew(at1,3);
    switch(izp_atnum[i]){
    case 1:
      sprintf(at1,"h");
	break;
    case 6:
      sprintf(at1,"c");
      break;
    case 7:
      sprintf(at1,"n");
      break;
    case 8:
      sprintf(at1,"o");
      break;
    case 9:
      sprintf(at1,"f");
      break;
    case 16:
      sprintf(at1,"s");
      break;
    case 30:
      sprintf(at1,"zn");
      break;
    default:
      gmx_fatal(FARGS,
		"no atomtype parameters for atom number %d\n",izp_atnum[i]);
    }
    /* get the second atom to make up the file
     */
    for(j=0;j<izp_nr;j++){
      snew(at2,3);
      switch(izp_atnum[j]){
      case 1:
	sprintf(at2,"h");
	break;
      case 6:
	sprintf(at2,"c");
	break;
      case 7:
	sprintf(at2,"n");
	break;
      case 8:
	sprintf(at2,"o");
	break;
      case 9:
	sprintf(at2,"f");
	break;
      case 16:
	sprintf(at2,"s");
	break;
      case 30:
	sprintf(at2,"zn");
	break;
      default:
	gmx_fatal(FARGS,
		  "no atomtype parameters for atom number %d\n",izp_atnum[j]);
      }
      //      snew(skf[k],2048); /* 500 characters should be enough */
      sprintf(skf,"%s'%s/%s%s.spl'\n",
	      skf,buf,at1,at2);
      free(at2);
    }
    free(at1);
  }
  tmpfile=fopen("slkf.inp","w");
  fprintf(tmpfile,"%s\n",skf);
  fclose(tmpfile);
  //  for(i=0;i<qm->nrQMatoms;i++)
  // fprintf(stderr,"izp[%d]= %d\n",i,izp[i]);
  //for(i=0;i<qm->nrQMatoms*3;i++)
  //  fprintf(stderr,"qmcrd[%d]= %lf\n",i,qmcrd[i]);
  scftol = (double)qm->scftol;
  fprintf(stderr,"calling gmxcao with scftol %lf\n",scftol);
  F77_FUNC(gmxcao,GMXCAO)(&scftol,&qm->QMcharge,&qm->nrQMatoms,izp,
			  qmcrd,lmax,skf,&telec,&aantal);

  //  fprintf(stderr,"aantal = %d\n",(int)aantal);
  
  qm->F77_arraysize = aantal;
  free(qmcrd);
  free(izp);
  free(izp_atnum);
  free(lmax);
  free(skf);
  fprintf(stderr,"DFTB initialized\n");
  

} /* init_mopac */

real call_dftb(t_commrec *cr, t_forcerec *fr, t_QMrec *qm, t_MMrec *mm, 
	       rvec f[], rvec fshift[])
{
  /* do the actual QMMM calculation using directly linked dftb subroutines 
   */
  double
    *qmcrd=NULL,*mmcrd=NULL,*mmchrg=NULL,
    *qmgrad=NULL,*mmgrad=NULL,energy,telec,scftol;
  int
    i,j,scc;
  real
    QMener=0.0;

  if(mm->nrMMatoms > qm->F77_arraysize){
    gmx_fatal(FARGS,"Nr. of MM atoms (%d) larger than arraysizes in DFTB (%d)\nrecomplile DFTB with more atoms\n",mm->nrMMatoms,qm->F77_arraysize);
  }
  if(qm->nrQMatoms > qm->F77_arraysize){
    gmx_fatal(FARGS,"Nr. of QM atoms (%d) larger than arraysizes in DFTB (%d)\nrecomplile DFTB with more atoms\n",qm->nrQMatoms,qm->F77_arraysize);
  }
  

  //  fprintf(stderr,"array size in dftb: %d\n",qm->F77_arraysize);
  snew(qmcrd, 3*(qm->F77_arraysize));
  snew(qmgrad,3*(qm->F77_arraysize));
  
  snew(mmcrd,3*(qm->F77_arraysize));
  snew(mmchrg,1*qm->F77_arraysize);
  snew(mmgrad,3*(qm->F77_arraysize));


  for(i=0;i<qm->nrQMatoms;i++){
    for (j=0;j<DIM;j++){
      qmcrd[3*i+j] = (double)qm->xQM[i][j]*10;
    }
  }
  for(i=0;i<mm->nrMMatoms;i++){
    for (j=0;j<DIM;j++){
      mmcrd[3*i+j] = (double)mm->xMM[i][j]*10;
    }
  
  /* charge is the last entry */
    mmchrg[i] = (double)mm->MMcharges[i];
  }
  telec = (double)qm->telec;
  scftol = (double)qm->scftol;

  F77_FUNC(dodftb,DODFTB)(&scftol,&qm->nrQMatoms,qmcrd,&mm->nrMMatoms,
  		  mmcrd,mmchrg,qmgrad,mmgrad,&telec,&energy);

  //  fprintf(stderr,"dftb called....energy is %lf\n",energy);
  /* add the gradients to the f[] array, and also to the fshift[].
   * the DFTB gradients are in ?????.
   */
  
  
  for(i=0;i<qm->nrQMatoms;i++){
      for(j=0;j<DIM;j++){
	  f[i][j]       = -qmgrad[3*i+j];
      }
      //      fprintf(stderr,"f[%d] = {%10.7lf,%10.7lf,%10.7lf}\n",
      //		  i,f[i][0],f[i][1],f[i][2]);

  }


  for(i=0;i<qm->nrQMatoms;i++){
    for(j=0;j<DIM;j++){
      f[i][j]       = (real)HARTREE_BOHR2MD*qmgrad[3*i+j];
      fshift[i][j]  = (real)HARTREE_BOHR2MD*qmgrad[3*i+j];
      //      fprintf(stderr,"f[%d] = {%10.7f,%10.7f,%10.7f}\n",
      //      i,f[i][0],f[i][1],f[i][2]);
    }
  }
  //  fprintf(stderr,"MM gradients:\n");
  for(i=0;i<mm->nrMMatoms;i++){
    for(j=0;j<DIM;j++){
      f[i+qm->nrQMatoms][j]      =  (real)HARTREE_BOHR2MD*mmgrad[3*i+j];      
      fshift[i+qm->nrQMatoms][j] =  (real)HARTREE_BOHR2MD*mmgrad[3*i+j];
      //      fprintf(stderr,"f[%d] = {%10.7f,%10.7f,%10.7f}\n",
      //      i,f[i+qm->nrQMatoms][0],f[i+qm->nrQMatoms][1],f[i+qm->nrQMatoms][2]);
    }
  }  
  QMener = (real)energy;
  //  fprintf(stderr,"QMener 1 = %f\n",QMener);
  QMener *= HARTREE2KJ*AVOGADRO;
  //  fprintf(stderr,"QMener 2 = %f\n",QMener);
  /* do we do something with the mulliken charges?? 
   */
  free(mmcrd);
  free(mmchrg);
  free(mmgrad);

  free(qmcrd);

  free(qmgrad);  
  //  fprintf(stderr,"hm, what goes wrong then....\n");
  return (QMener);
}


